from typing import MutableMapping, Any, TYPE_CHECKING, List, Dict

from pydantic import PrivateAttr, BaseModel

if TYPE_CHECKING:
    from config_wrangler.config_from_loaders import ConfigFromLoaders


class ConfigHierarchy(BaseModel):
    """
    NOTE: This class requires that the top of the hierarchy be an instance of ConfigFromLoaders
    """
    _root_config: 'ConfigFromLoaders' = PrivateAttr(default=None)
    _parents: List[str] = PrivateAttr()
    _name_map: Dict[str, str] = PrivateAttr(default={})

    # noinspection PyMethodParameters
    def __init__(__pydantic_self__, **data: Any) -> None:
        """
        Create a new model by parsing and validating input data from keyword arguments.

        Raises ValidationError if the input data cannot be parsed to form a valid model.

        Uses something other than `self` the first arg to allow "self" as a settable attribute
        """
        super().__init__(**data)

    def full_item_name(self, item_name: str = None, delimiter: str = ' -> '):
        try:
            if item_name is None:
                return delimiter.join(self._parents)
            else:
                return delimiter.join(self._parents + [item_name])
        except AttributeError as e:
            raise AttributeError(f"{e} not found in {repr(self)}")

    @staticmethod
    def translate_config_data(config_data: MutableMapping):
        return config_data

    def _translate_name(self, old_name):
        if old_name in self._name_map:
            return self._name_map[old_name]
        else:
            return old_name

    def get(self, section, item, fallback=...):
        try:
            section_obj = getattr(self, self._translate_name(section))
            return getattr(section_obj, item)
        except AttributeError:
            if fallback is ...:
                raise
            else:
                return fallback

    def getboolean(self, section, item, fallback=...) -> bool:
        value = self.get(section=section, item=item, fallback=fallback)
        if value is None:
            value = False
        if not isinstance(value, bool):
            raise ValueError('getboolean called on non-bool config item')
        return value

    def get_list(self, section, item, fallback=...) -> list:
        value = self.get(section=section, item=item, fallback=fallback)
        if value is None:
            value = []
        if not isinstance(value, list):
            raise ValueError('get_list called on non-list config item')
        return value

    def __getitem__(self, section):
        try:
            section_obj = getattr(self, section)
            return section_obj.dict()
        except AttributeError as e:
            raise KeyError(str(e))



import inspect
import logging
import os
import typing
from copy import deepcopy

from pydantic import BaseModel

from config_wrangler.config_data_loaders.base_config_data_loader import BaseConfigDataLoader
from config_wrangler.config_exception import ConfigError
from config_wrangler.config_types.dynamically_referenced import DynamicallyReferenced
from config_wrangler.utils import walk_model, match_config_data_to_field


class EnvConfigDataLoader(BaseConfigDataLoader):
    """
    Copied initial version from Pydantic BaseSettings.
    Made to fit into ConfigLoader approach with nested models.
    """

    def __init__(self, env_prefix: str = None, config_data_dict: typing.Dict[str, typing.Any] = None, **kwargs):
        super().__init__(config_data_dict=config_data_dict, **kwargs)
        self.env_prefix = env_prefix or ''
        self.log = logging.getLogger(__name__)

    def read_config_data(self, model: BaseModel) -> typing.MutableMapping:
        config_data = deepcopy(self._init_config_data)

        env_vars = {k.lower(): v for k, v in os.environ.items()}

        for field, parents in walk_model(model):
            env_val: typing.Optional[str] = None
            # Note: field.field_info.extra['env_names'] is set from the 'env' variable on the field or in Config.
            #       See https://pydantic-docs.helpmanual.io/usage/settings/#environment-variable-names
            if 'env_names' in field.field_info.extra:
                for env_name in field.field_info.extra['env_names']:
                    env_name = self.env_prefix + env_name
                    env_val = env_vars.get(env_name.lower())
                    if env_val is not None:
                        break

            dynamic_ref = False
            if inspect.isclass(field.type_) and issubclass(field.type_, DynamicallyReferenced):
                dynamic_ref = True
            possible_names = ('_', '.', ':')

            if env_val is None:
                possible_env_vars = set()

                for field_name in [field.alias, field.name]:
                    for delimiter in possible_names:
                        if field_name:
                            full_name = delimiter.join(parents + [field_name])
                        else:
                            full_name = delimiter.join(parents)
                        env_name = f"{self.env_prefix}{full_name}"
                        if env_name in env_vars:
                            possible_env_vars.add(env_name)
                if len(possible_env_vars) > 1:
                    raise ValueError(f"Found multiple matches for {parents} {field}.  They are: {possible_env_vars}")
                elif len(possible_env_vars) == 1:
                    env_var = list(possible_env_vars)[0]
                    # TODO: Find a way to log this later since logging is probably not setup yet
                    self.log.debug(f"Read ENV {env_var} into {parents} {field}.")
                    env_val = env_vars[env_var]

            if field.is_complex():
                env_val = match_config_data_to_field(
                    field=field,
                    field_value=env_val,
                    create_from_section_names=False,  # Not yet supported for env
                    parent_container={},
                    root_config_data={},
                    parents=parents
                )

            # Only set the value if it is not already in our config_data
            if env_val is not None:
                sub_config = config_data
                for parent in parents:
                    if parent in sub_config:
                        sub_config = sub_config[parent]
                    else:
                        new_sub_config = dict()
                        sub_config[parent] = new_sub_config
                        sub_config = new_sub_config
                if field.alias not in sub_config:
                    sub_config[field.alias] = env_val
        return config_data




import os
from pathlib import Path
from pydantic import DirectoryPath
from pydantic.validators import path_validator

__all__ = ['DirectoryPath', 'AutoCreateDirectoryPath', 'DirectoryFindUp', 'PathExpandUser']


class PathExpandUser(DirectoryPath):
    @staticmethod
    def _expand_user(path: Path):
        path = path.expanduser()
        return path

    @classmethod
    def __get_validators__(cls):
        yield path_validator
        yield cls._expand_user
        super().__get_validators__()


class AutoCreateDirectoryPath(PathExpandUser):
    @staticmethod
    def _ensure_exsits(path: Path):
        if not path.exists():
            os.makedirs(path)
        return path

    @classmethod
    def __get_validators__(cls):
        yield path_validator
        yield cls._expand_user
        yield cls._ensure_exsits
        super().__get_validators__()


class DirectoryFindUp(DirectoryPath):
    @staticmethod
    def __find_up(path: Path):
        if path.exists():
            return path
        else:
            start_dir = Path(os.getcwd())
            for parent_dir in start_dir.parents:
                parent_path = Path(parent_dir, path)
                if parent_path.exists():
                    return parent_path
            raise FileNotFoundError(f"{path} not found in {start_dir} or parents")

    @classmethod
    def __get_validators__(cls):
        yield path_validator
        yield cls.__find_up
        yield cls.validate



import logging
import typing

from pydantic import BaseModel, PrivateAttr

from config_wrangler.config_data_loaders.base_config_data_loader import BaseConfigDataLoader
from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy
from config_wrangler.config_templates.credentials import PasswordDefaults
from config_wrangler.utils import merge_configs, interpolate_values


class ConfigFromLoaders(ConfigHierarchy):
    """
    Base class for settings, allowing values to be set by files or environment variables.
    """
    _model_validators: PrivateAttr(default=[])
    passwords: PasswordDefaults = None

    # noinspection PyMethodParameters
    def __init__(
        __pydantic_self__,
        _config_data_loaders: typing.List[BaseConfigDataLoader],
        **kwargs: typing.Dict[str, typing.Any]
    ) -> None:
        """
        Note: Uses something other than `self` the first arg to allow "self" as a settable attribute
        """
        log = logging.getLogger(__name__)

        config_data = dict(**kwargs)
        for loader in _config_data_loaders:
            log.debug(f"Loading config with {loader}")
            loader_config_data = loader.read_config_data(__pydantic_self__)
            merge_configs(config_data, loader_config_data)
        log.debug("Interpolating config macro references")
        interpolate_values(config_data, config_data)
        log.debug("Translating config with translate_config_data method")
        config_data = __pydantic_self__.translate_config_data(config_data)
        log.debug(f"Calling pydantic __init__")
        super().__init__(**config_data)
        log.debug("Calling fill_hierarchy to fill in root and parent data")
        __pydantic_self__.validate_model()

    def fill_hierarchy_any_type(
            self,
            value: object,
            parents: typing.List[str],
            errors: set,
    ):
        if isinstance(value, BaseModel):
            self.fill_hierarchy(
                model_level=value,
                parents=parents,
                errors=errors
            )
        elif isinstance(value, list):
            for index, entry in enumerate(value):
                self.fill_hierarchy_any_type(
                    value=entry,
                    parents=parents + [f"[{index}]"],
                    errors=errors
                )
        elif isinstance(value, dict):
            for key, entry in value.items():
                self.fill_hierarchy_any_type(
                    value=entry,
                    parents=parents + [f"[{key}]"],
                    errors=errors
                )

    def fill_hierarchy(
            self,
            model_level: BaseModel,
            parents: typing.List[str],
            errors: set,
    ):
        log = logging.getLogger(__name__)
        if len(parents) > 10:
            raise ValueError(f"Possible model self reference {parents}")
        try:
            model_level._root_config = self
            model_level._parents = parents
            # noinspection PyUnresolvedReferences
            name = model_level.full_item_name()
            log.debug(f"fill_hierarchy on {name}")
        except AttributeError as e:
            log.warning(f"{parents} {repr(model_level)} is not an instance inheriting from ConfigHierarchy: {e}")
        for attr_name, attr_value in model_level.__dict__.items():
            self.fill_hierarchy_any_type(
                value=attr_value,
                parents=parents + [attr_name],
                errors=errors
            )

        # Run any model level validators
        for attr in dir(model_level):
            if attr.startswith('_validate_model_'):
                method = getattr(model_level, attr)
                if callable(method):
                    try:
                        method()
                    except (ValueError, TypeError, AssertionError) as exc:
                        errors.add(repr(exc))

    def validate_model(self):
        errors = set()
        self.fill_hierarchy(
            model_level=self,
            parents=[],
            errors=errors,
        )
        if len(errors) > 0:
            log = logging.getLogger(__name__)
            log.error(f"{len(errors)} config errors found:")
            for error in errors:
                log.error(error)
            indent = ' ' * 3
            errors_str = f"\n{indent}".join(errors)
            raise ValueError(f"Config Errors (cnt={len(errors)}). Errors=\n{indent}{errors_str}")


import ast
import inspect
import json
import logging
import re
import typing
from datetime import timezone, datetime

from pydantic import BaseModel
from pydantic.fields import SHAPE_LIST, SHAPE_SINGLETON, SHAPE_TUPLE, SHAPE_ITERABLE, SHAPE_SEQUENCE, \
    SHAPE_TUPLE_ELLIPSIS, SHAPE_SET, SHAPE_FROZENSET, SHAPE_DICT, SHAPE_DEFAULTDICT, Field, ModelField
from pydantic.utils import lenient_issubclass

from config_wrangler.config_exception import ConfigError
from config_wrangler.config_types.dynamically_referenced import DynamicallyReferenced


class TZFormatter(logging.Formatter):
    local_timezone = datetime.now(timezone.utc).astimezone().tzinfo

    @staticmethod
    def tz_aware_converter(timestamp) -> datetime:
        return datetime.fromtimestamp(timestamp, TZFormatter.local_timezone)

    def formatTime(self, record, datefmt=None):
        dt = TZFormatter.tz_aware_converter(record.created)
        if datefmt is not None:
            result = dt.strftime(datefmt)
        else:
            time_part = dt.strftime(self.default_time_format)
            result = self.default_msec_format % (time_part, record.msecs)
        return result


def merge_configs(child: typing.MutableMapping, parent: typing.MutableMapping) -> None:
    for section in parent:
        if section not in child:
            child[section] = parent[section]
        else:
            if isinstance(child[section], typing.MutableMapping):
                merge_configs(child[section], parent[section])
            else:
                pass


def resolve_variable(root_config_data: typing.MutableMapping, variable_name: str, default=None, part_delimiter=':') -> typing.Any:
    variable_name_parts = variable_name.split(part_delimiter)
    result = root_config_data
    for part in variable_name_parts:
        if part in result:
            result = result[part]
        else:
            return default
    return result


_interpolation_re = re.compile(r"\${([^}]+)}")


def interpolate_values(container: typing.MutableMapping, root_config_data: typing.MutableMapping):
    for section in container:
        value = container[section]
        if isinstance(value, typing.MutableMapping):
            interpolate_values(value, root_config_data=root_config_data)
        elif isinstance(value, str):
            if '$' in value:
                variables_cnt = 0
                result_values = []
                next_start = 0
                for variable_found in _interpolation_re.finditer(value):
                    variables_cnt += 1
                    variable_name = variable_found.groups()[0]
                    variable_text = f"${{{variable_name}}}"
                    var_start, var_end = variable_found.span()

                    if ':' in variable_name:
                        variable_replacement = resolve_variable(
                            root_config_data, variable_name, default=variable_text, part_delimiter=':'
                        )
                    elif '.' in variable_name:
                        variable_replacement = resolve_variable(
                            root_config_data, variable_name, default=variable_text, part_delimiter='.'
                        )
                    else:
                        variable_replacement = container.get(variable_name, default=variable_text)

                    result_values.append(value[next_start:var_start])
                    result_values.append(variable_replacement)
                    next_start = var_end + 1
                if variables_cnt > 0:
                    if next_start < len(value):
                        result_values.append(value[next_start:])
                    container[section] = ''.join(result_values)


def parse_as_literal_or_json(value: str) -> typing.Any:
    try:
        return ast.literal_eval(value)
    except (ValueError, SyntaxError) as el:
        try:
            return json.loads(value)
        except ValueError as ej:
            raise ConfigError(
                f"Value {value} could not be parsed as python literal '{el}' or json '{ej}'"
            )


def parse_delimited_list(field: Field, value: str) -> typing.Sequence:
    value = value.strip()
    if len(value) == 0:
        return []
    delimiter = field.field_info.extra.get('delimiter', None)
    if delimiter is None and value[0] not in {'[', '{'}:
        # Try to automatically recognize the delimiter
        for delimiter in ['\n', ',', '|']:
            if delimiter in value:
                break
    if delimiter is not None:
        if value[0] == delimiter:
            value = value[1:]
        result = [v.strip() for v in value.split(delimiter)]
        if field.type_ == int:
            result = [int(v) for v in result]
        elif field.type_ == float:
            result = [float(v) for v in result]
    else:
        try:
            result = parse_as_literal_or_json(value)
        except ValueError as e:
            raise ConfigError(f"Field {field.alias} {e}")
    return result


def full_name(parents: typing.List[str], field_name: str):
    return '.'.join(parents + [field_name])


def inherit_fill(parent_config, child_config):
    for inherit_key, inherit_value in parent_config.items():
        if inherit_key not in child_config:
            child_config[inherit_key] = inherit_value


def find_referenced_section(
        field: ModelField,
        parents: typing.List[str],
        section_name: typing.Union[str, typing.MutableMapping],
        current_dict: typing.MutableMapping,
        root_dict: typing.MutableMapping
) -> typing.MutableMapping:
    inherit = field.field_info.extra.get('inherit', False)
    if isinstance(section_name, str):
        if section_name in root_dict:
            if inherit:
                inherit_fill(current_dict, root_dict[section_name])
            section_value = root_dict[section_name]
        elif section_name in current_dict:
            if inherit:
                inherit_fill(current_dict, current_dict[section_name])
            section_value = current_dict[section_name]
        else:
            raise ConfigError(
                f"{full_name(parents, field.alias)} refers to section {section_name} which does not exist."
            )
        return section_value
    else:
        return section_name


def match_config_data_to_field(
        field: ModelField,
        field_value: object,
        create_from_section_names: bool,
        parent_container: typing.MutableMapping,
        root_config_data: typing.MutableMapping,
        parents: typing.List[str],
):
    if field.shape == SHAPE_SINGLETON and field.type_ in {int, float}:
        # if field.type_ == int:
        #     field_value = int(str(field_value))
        # elif field.type_ == float:
        #     field_value = float(str(field_value))
        pass
    elif (
        field.shape in {SHAPE_LIST, SHAPE_TUPLE, SHAPE_TUPLE_ELLIPSIS, SHAPE_ITERABLE, SHAPE_SEQUENCE}
        or (field.shape == SHAPE_SINGLETON and field.type_ in {list, tuple})
    ):
        if isinstance(field_value, str):
            field_value = parse_delimited_list(field, field_value)

        # Conditional change to field_value to obtain value from another config section
        if create_from_section_names:
            if not hasattr(field.type_, '__fields__'):
                raise ValueError(f"{full_name(parents, field.alias)} has create_from_section_names but has no fields")

            section_contents = list()
            for section_name in field_value:
                section_value = find_referenced_section(
                    field=field,
                    parents=parents,
                    section_name=section_name,
                    current_dict=parent_container,
                    root_dict=root_config_data,
                )
                if field.sub_fields:
                    # Transform section_value as needed
                    sub_field = field.sub_fields[0]
                    section_value = match_config_data_to_field_or_submodel(
                        field=sub_field,
                        parent_container={sub_field.alias: section_value},
                        root_config_data=root_config_data,
                        parents=parents + [field.alias, section_name]
                    )
                # Add name to the dict in case the model wants it
                section_value['config_source_name'] = section_name
                section_contents.append(section_value)
            field_value = section_contents
    elif (
            field.shape in {SHAPE_DICT, SHAPE_DEFAULTDICT}
            or (field.shape == SHAPE_SINGLETON and field.type_ in {dict})
    ):
        if isinstance(field_value, str):
            try:
                field_value = parse_as_literal_or_json(field_value)
            except ValueError as e:
                raise ConfigError(f"Field {full_name(parents, field.alias)} {e}")

        # Conditional change to field_value to obtain value from another config section
        if create_from_section_names:
            section_contents = dict()
            for key, value in field_value.items():
                if field.sub_fields:
                    sub_field = field.sub_fields[0]
                    sub_field.field_info.extra['create_from_section_names'] = True
                    # Transform section_value as needed
                    section_value = match_config_data_to_field(
                        field=sub_field,
                        field_value=value,
                        create_from_section_names=True,
                        parent_container=field_value,
                        root_config_data=root_config_data,
                        parents=parents + [field.alias, key, value]
                    )
                    section_contents[key] = section_value
                else:
                    section_value = find_referenced_section(
                        field=field,
                        parents=parents,
                        section_name=value,
                        current_dict=parent_container,
                        root_dict=root_config_data,
                    )
                    section_contents[key] = section_value
            field_value = section_contents
    elif (
            field.shape in {SHAPE_SET}
            or (field.shape == SHAPE_SINGLETON and field.type_ in {set})
    ):
        if isinstance(field_value, str):
            field_value = set(parse_delimited_list(field, field_value))
    elif (
            field.shape in {SHAPE_FROZENSET}
            or (field.shape == SHAPE_SINGLETON and field.type_ in {frozenset})
    ):
        if isinstance(field_value, str):
            field_value = frozenset(parse_delimited_list(field, field_value))
    # In all cases not explicitly matched above, we'll let pydantic parse it as is
    return field_value


def match_config_data_to_model(
        model: BaseModel,
        config_data: typing.MutableMapping,
        root_config_data: typing.MutableMapping = None,
        parents=None
):
    if parents is None:
        parents = []
    if root_config_data is None:
        root_config_data = config_data

    # Make mappings from lower case names to actual config field names
    config_name_map = {key.lower(): key for key in config_data}
    root_config_name_map = {key.lower(): key for key in root_config_data}

    # Scan all model fields to look for values in the MutableMapping
    for field in model.__fields__.values():
        # Check for either a direct name match or a case in-sensitive match
        field_lower = field.alias.lower()

        if field.alias in config_data:
            found = True
        elif field_lower in config_name_map:
            found = True
            field_name = config_name_map[field_lower]
            config_data[field.alias] = config_data[field_name]
            del config_data[field_name]
        else:
            found = False

        # Check for nested objects set using top level dotted names (e.g. [parent.child])
        # (either a direct name match or a case in-sensitive match)
        if not found and len(parents) > 0:
            section_name = '.'.join(parents + [field.alias])
            section_name_lower = section_name.lower()
            if section_name in root_config_data:
                found = True
                # Copy data into place where pydantic will expect it
                config_data[field.alias] = root_config_data[section_name]
            elif section_name_lower in root_config_name_map:
                found = True
                section_name = root_config_name_map[section_name_lower]
                # Copy data into place where pydantic will expect it
                config_data[field.alias] = root_config_data[section_name]

        if found:
            updated_value = match_config_data_to_field_or_submodel(
                field=field,
                parent_container=config_data,
                root_config_data=root_config_data,
                parents=parents + [field.alias]
            )
            config_data[field.alias] = updated_value
    return config_data


def match_config_data_to_field_or_submodel(
        field: ModelField,
        parent_container: typing.MutableMapping,
        root_config_data: typing.MutableMapping = None,
        parents=None
):
    create_from_section_names = field.field_info.extra.get('create_from_section_names', False)

    if inspect.isclass(field.type_) and issubclass(field.type_, DynamicallyReferenced):
        updated_value = match_config_data_to_field(
            field=field,
            field_value=parent_container[field.alias],
            create_from_section_names=create_from_section_names,
            parent_container=parent_container,
            root_config_data=root_config_data,
            parents=parents
        )
        if isinstance(updated_value, list):
            updated_value = [{'ref': entry} for entry in updated_value]
        else:  # Dict
            updated_value = {key: {'ref': entry} for key, entry in updated_value.items()}
    elif hasattr(field.type_, '__fields__') and not create_from_section_names:
        updated_value = match_config_data_to_model(
            model=field.type_,
            config_data=parent_container[field.alias],
            root_config_data=root_config_data,
            parents=parents
        )
    else:
        updated_value = match_config_data_to_field(
            field=field,
            field_value=parent_container[field.alias],
            create_from_section_names=create_from_section_names,
            parent_container=parent_container,
            root_config_data=root_config_data,
            parents=parents
        )
    return updated_value


def walk_model(
        model: BaseModel,
        parents=None
):
    if parents is None:
        parents = []

    # Scan all model fields to look for values in the MutableMapping
    for field in model.__fields__.values():
        if inspect.isclass(field.type_) and issubclass(field.type_, DynamicallyReferenced):
            yield field, parents
        elif hasattr(field.type_, '__fields__'):
            yield from walk_model(
                model=field.type_,
                parents=parents + [field.alias]
            )
        else:
            yield field, parents



from pydantic import validator

from config_wrangler.config_templates.config_hierarchy import ConfigHierarchy


class DynamicallyReferenced(ConfigHierarchy):
    ref: str

    @validator('ref')
    def _validate_phase_1(cls, value):
        if value == '':
            raise ValueError('Blank is not valid for a DynamicallyReferenced section')
        return value

    def _validate_model_reference(self):
        _ = self.get_referenced()

    def get_referenced(self) -> ConfigHierarchy:
        if self.ref is None:
            raise ValueError(f"DynamicallyReferenced {self} is not set")
        elif self._root_config is None:
            raise RuntimeError(
                "DynamicallyReferenced._root_config is none. "
                " Either the root class is not ConfigFromLoaders or "
                "ConfigFromLoaders.fill_hierarchy did not work."
            )
        else:
            parts = self.ref.split('.')

            model = self._root_config
            for part in parts:
                try:
                    model = getattr(model, part)
                except AttributeError:
                    raise ValueError(f"{self.ref} not found.")
            return model

    def __str__(self):
        return f"{self.ref}"

    def __repr__(self):
        return f"{self.__class__.__name__}({self.ref}) where get_referenced returns {self.get_referenced().__class__.__name__} instance"



import logging
import typing

from pydantic import BaseModel


class BaseConfigDataLoader:
    def __init__(self, config_data_dict: typing.Dict[str, typing.Any] = None, **kwargs):
        self.log = logging.getLogger(f"{self.__class__.__module__}.{self.__class__.__name__}")
        if config_data_dict is None:
            self._init_config_data = dict(**kwargs)
        else:
            self._init_config_data = dict(**kwargs).update(config_data_dict)

    def _apply_types(self, config_data, model: BaseModel) -> typing.MutableMapping:
        for field in model.__fields__.values():
            self.log.debug(field.alias)
            # if field.is_complex():
            #     raise NotImplementedError
            # else:
            #     config_data[field.alias] =
        return config_data

    def read_config_data(self, model: BaseModel) -> typing.MutableMapping:
        return self._apply_types(self._init_config_data)

    def save_config_data(self, config_data: BaseModel):
        raise RuntimeError(f"{self.__class__.__name__} doesn't deal with files")



